/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include "nxpcanvasobject.h"
#include "nxpcanvasbbox.h"
#include <math.h>
#include <string.h>
#include "nxplot-marshalers.h"
#include "nxplot-artft.h"
#include "nxpcanvasmisc.h"


#define WBBOX_BORDER 1

enum {
    DRAW_SIGNAL,
    UPDATE_SIGNAL,
    ENTER_SIGNAL,
    LEAVE_SIGNAL,
    BUTTON_PRESS_SIGNAL,
    BUTTON_RELEASE_SIGNAL,
    MOTION_NOTIFY_SIGNAL,
    SIZE_REQUEST_SIGNAL,
    CONFIGURE_SIGNAL,
    KEY_PRESS_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 }; 

enum {
    PROP_0,
    PROP_CANVAS,
    PROP_INTER_MASK,
    PROP_WHOLE_AREA_INTERACTION,
};

static gpointer parent_class = NULL;

NxpCanvasObject* nxp_canvas_object_new(void);

static void nxp_canvas_object_init                   (NxpCanvasObject            *instance,
						      gpointer                    g_class);
static void nxp_canvas_object_class_base_init        (NxpCanvasObjectClass       *class);
static void nxp_canvas_object_class_init             (NxpCanvasObjectClass       *class,
						      gpointer                    data);
static void nxp_canvas_object_shutdown               (GObject                    *object);
static void set_property                             (GObject                    *object,
						      guint                       prop_id,
						      const GValue               *value,
						      GParamSpec                 *pspec);
static void get_property                             (GObject                    *object,
						      guint                       prop_id,
						      GValue                     *value,
						      GParamSpec                 *pspec);
void        nxp_canvas_object_set_interaction_cursor (NxpCanvasObject            *self,
						      NxpCanvasObjectInteraction  inter);
void        nxp_canvas_object_get_interaction_bbox   (NxpCanvasObject            *self,
						      NxpCanvasObjectInteraction  inter,
						      gint                        x,
						      gint                        y,
						      GdkRectangle               *new_bbox);
static void nxp_canvas_object_real_enter             (NxpCanvasObject            *self);
static void nxp_canvas_object_real_leave             (NxpCanvasObject            *self);
static gboolean nxp_canvas_object_real_button_press      (NxpCanvasObject            *self,
						      GdkEvent                   *event);
static gboolean nxp_canvas_object_real_button_release    (NxpCanvasObject            *self,
						      GdkEvent                   *event);
static gboolean nxp_canvas_object_real_motion_notify     (NxpCanvasObject            *self,
						      GdkEvent                   *event);


NxpCanvasObjectInteraction nxp_canvas_object_get_interaction_at_point (NxpCanvasObject *self,
								       gint             x,
								       gint             y);




static GTypeInfo typeinfo = 
{
    /* interface types, classed types, instantiated types */
    sizeof(NxpCanvasObjectClass), 

    (GBaseInitFunc)nxp_canvas_object_class_base_init, 
    NULL, // GBaseFinalizeFunc      base_finalize;

    /* classed types, instantiated types */
    (GClassInitFunc)nxp_canvas_object_class_init, 
    NULL, // GClassFinalizeFunc     class_finalize;
    NULL, // gconstpointer          class_data;

    /* instantiated types */
    sizeof(NxpCanvasObject),		//guint16                instance_size;
    0,					//guint16                n_preallocs;
    (GInstanceInitFunc)nxp_canvas_object_init, //GInstanceInitFunc      instance_init;

    /* value handling */
    NULL,				//const GTypeValueTable	*value_table;
};


GType nxp_canvas_object_get_type(void)
{
    static GType type = 0;
    if (!type)
    {
	type = g_type_register_static(G_TYPE_OBJECT, "NxpCanvasObject",
				      &typeinfo, 0);
    }
    return type;
}


static void nxp_canvas_object_init(NxpCanvasObject *self, gpointer g_class)
{
/*     g_message(__FUNCTION__); */
    /* the real bbox geometry will be set in the update_bbox method */
    self->bbox_overlay = g_object_new(NXP_TYPE_CANVAS_BBOX,
				      "line-style", GDK_LINE_ON_OFF_DASH,
				      "handles", self->interaction_mask,
				      NULL);
    if (self->canvas)
	nxp_canvas_overlay_set_canvas(self->bbox_overlay, self->canvas);
    self->ghost_bbox = g_object_new(NXP_TYPE_CANVAS_BBOX,
				    "line-style", GDK_LINE_ON_OFF_DASH,
				    "handles", 0,
				    NULL);
    if (self->canvas)
	nxp_canvas_overlay_set_canvas(self->ghost_bbox, self->canvas);
    self->interaction = NXP_COBJ_INTERACTION_NONE;
    self->dirty = FALSE;
    self->visible = TRUE;
}


static void nxp_canvas_object_class_base_init(NxpCanvasObjectClass *class)
{
    /* empty */
}


static void nxp_canvas_object_class_init(NxpCanvasObjectClass *class, gpointer data)
{
    GObjectClass *object_class = G_OBJECT_CLASS(class);

    object_class->set_property = set_property;
    object_class->get_property = get_property;
    object_class->finalize     = nxp_canvas_object_shutdown;
    
    class->enter	  = nxp_canvas_object_real_enter;
    class->leave	  = nxp_canvas_object_real_leave;
    class->motion_notify  = nxp_canvas_object_real_motion_notify;
    class->button_press   = nxp_canvas_object_real_button_press;
    class->button_release = nxp_canvas_object_real_button_release;

    parent_class = g_type_class_peek_parent (class);
    

    g_object_class_install_property(object_class,
				    PROP_CANVAS,
				    g_param_spec_object(
					"canvas", "Parent NxpCanvas",
					"Object's parent NxpCanvas",
					NXP_TYPE_CANVAS,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_INTER_MASK,
				    g_param_spec_int(
					"interaction_mask",
					"NxpCanvasObjectInteractionMask",
					"Bit mask specifying allowed"
					" interactions with object",
					0, 0xffffff, 0,
					G_PARAM_READABLE|G_PARAM_WRITABLE|
					G_PARAM_CONSTRUCT));
    g_object_class_install_property(object_class,
				    PROP_WHOLE_AREA_INTERACTION,
				    g_param_spec_boolean(
					"whole_area_interaction",
					"whole_area_interaction",
					"Whether the whole bounding box of"
					" the object is available as"
					" interaction area", FALSE,
					G_PARAM_WRITABLE|G_PARAM_CONSTRUCT));

    signals[DRAW_SIGNAL] = 
	g_signal_new("draw",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, draw),
		     NULL, NULL,
		     _nxplot_marshal_VOID__OBJECT,
		     G_TYPE_NONE, 1, TYPE_NXPLOT);
    signals[UPDATE_SIGNAL] = 
	g_signal_new("update",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, update),
		     NULL, NULL,
		     _nxplot_marshal_VOID__VOID,
		     G_TYPE_NONE, 0);
    signals[ENTER_SIGNAL] = 
	g_signal_new("enter",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, enter),
		     NULL, NULL,
		     _nxplot_marshal_VOID__VOID,
		     G_TYPE_NONE, 0);
    signals[LEAVE_SIGNAL] = 
	g_signal_new("leave",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, leave),
		     NULL, NULL,
		     _nxplot_marshal_VOID__VOID,
		     G_TYPE_NONE, 0);
    signals[BUTTON_PRESS_SIGNAL] = 
	g_signal_new("button_press",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, button_press),
		     _nxp_boolean_handled_accumulator, NULL,
		     _nxplot_marshal_BOOLEAN__BOXED,
		     G_TYPE_BOOLEAN, 1, GDK_TYPE_EVENT);
    signals[BUTTON_RELEASE_SIGNAL] = 
	g_signal_new("button_release",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, button_release),
		     _nxp_boolean_handled_accumulator, NULL,
		     _nxplot_marshal_BOOLEAN__BOXED,
		     G_TYPE_BOOLEAN, 1, GDK_TYPE_EVENT);
    signals[MOTION_NOTIFY_SIGNAL] = 
	g_signal_new("motion_notify",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, motion_notify),
		     _nxp_boolean_handled_accumulator, NULL,
		     _nxplot_marshal_BOOLEAN__BOXED,
		     G_TYPE_BOOLEAN, 1, GDK_TYPE_EVENT);
    signals[SIZE_REQUEST_SIGNAL] = 
	g_signal_new("size-request",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, size_request),
		     NULL, NULL,
		     _nxplot_marshal_BOXED__VOID,
		     TYPE_NXPLOT_POINT, 0);
    signals[CONFIGURE_SIGNAL] = 
	g_signal_new("configure",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, configure),
		     NULL, NULL,
		     _nxplot_marshal_VOID__BOXED_INT,
		     G_TYPE_NONE, 2, TYPE_NXPLOT_RECT, G_TYPE_INT);
    signals[KEY_PRESS_SIGNAL] = 
	g_signal_new("key_press",
		     G_OBJECT_CLASS_TYPE(object_class),
		     G_SIGNAL_RUN_LAST,
		     G_STRUCT_OFFSET(NxpCanvasObjectClass, key_press),
		     _nxp_boolean_handled_accumulator, NULL,
		     _nxplot_marshal_BOOLEAN__BOXED,
		     G_TYPE_BOOLEAN, 1, GDK_TYPE_EVENT);
}


static void nxp_canvas_object_shutdown(GObject *object)
{
    NxpCanvasObject *self;
    self = NXP_CANVAS_OBJECT(object);
    if (self->bbox_overlay) g_object_unref(self->bbox_overlay);
    if (self->ghost_bbox)   g_object_unref(self->ghost_bbox);

    if (self->canvas_notify_sig)
	g_signal_handler_disconnect(self->container, self->canvas_notify_sig);
    self->canvas_notify_sig = 0;

    if (G_OBJECT_CLASS(parent_class)->finalize)
	G_OBJECT_CLASS(parent_class)->finalize(object);
}



void nxp_canvas_object_draw(NxpCanvasObject *self, Nxplot *pl)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    g_signal_emit(G_OBJECT(self), signals[DRAW_SIGNAL], 0, pl);
}


void nxp_canvas_object_update_wbbox(NxpCanvasObject *self)
{
    double wx1 = 0, wy1 = 0, wx2 = 0, wy2 = 0, t;

    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    g_return_if_fail(self->canvas);

    nxplot_space2window(self->canvas->plotter, self->bbox.x1, self->bbox.y1, &wx1, &wy1);
    nxplot_space2window(self->canvas->plotter, self->bbox.x2, self->bbox.y2, &wx2, &wy2);
    if (wx1 > wx2) {
	t = wx1;
	wx1 = wx2;
	wx2 = t;
    }
    if (wy1 > wy2) {
	t = wy1;
	wy1 = wy2;
	wy2 = t;
    }
    self->wbbox.x = floor(wx1);
    self->wbbox.y = floor(wy1);
    self->wbbox.width = ceil(wx2 - wx1);
    self->wbbox.height = ceil(wy2 - wy1);
    self->wbbox.x -= WBBOX_BORDER;
    self->wbbox.y -= WBBOX_BORDER;
    self->wbbox.width  += WBBOX_BORDER*2;
    self->wbbox.height += WBBOX_BORDER*2;
    if (self->bbox_overlay)
	nxp_canvas_bbox_set_geometry(NXP_CANVAS_BBOX(self->bbox_overlay),
				     &self->wbbox);
}


void nxp_canvas_object_enter(NxpCanvasObject *self)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    g_signal_emit(G_OBJECT(self), signals[ENTER_SIGNAL], 0);
}


static void nxp_canvas_object_real_enter(NxpCanvasObject *self)
{
/*     g_message(__FUNCTION__); */
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    if (self->bbox_overlay && self->interaction_mask) {
/* 	GdkRectangle *r = &self->bbox_overlay->bbox; */
/* 	g_message("Showing bbox overlay: (%i, %i, %i, %i); canvas=%p; im=%i", */
/* 		  r->x, r->y, r->width, r->height, */
/* 		  self->bbox_overlay->canvas, */
/* 		  self->interaction_mask); */
	nxp_canvas_overlay_show(self->bbox_overlay);
    }
}



void nxp_canvas_object_leave(NxpCanvasObject *self)
{
/*     NxpCanvasObjectClass *class; */
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    g_signal_emit(G_OBJECT(self), signals[LEAVE_SIGNAL], 0);
}


static void nxp_canvas_object_real_leave(NxpCanvasObject *self)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    if (self->bbox_overlay) {
/* 	g_message("Hiding bbox overlay"); */
	nxp_canvas_overlay_hide(self->bbox_overlay);
    }
    nxp_canvas_set_cursor(self->canvas, NULL);
}


void nxp_canvas_object_queue_draw(NxpCanvasObject *self)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    if (!self->canvas) return;
    if (!self->canvas->plotter) return;
    if (g_type_is_a(G_OBJECT_TYPE(self->canvas->plotter), TYPE_NXPLOT_ARTFT))
	nxp_canvas_redraw_area(self->canvas, &self->bbox);
}


static void
set_property (GObject      *object,
	      guint         prop_id,
	      const GValue *value,
	      GParamSpec   *pspec)
{
    NxpCanvasObject *self;

    self = NXP_CANVAS_OBJECT(object);

    switch (prop_id)
    {

    case PROP_CANVAS: {
	gpointer canv;
	NxpCanvas *canvas;

	g_return_if_fail(!self->canvas);
	canv = g_value_get_object(value);
	if (!canv) return;
	canvas = NXP_CANVAS(canv);
	if (self->canvas != canvas)
	    nxp_canvas_object_set_canvas(self, canvas);
	break;
    }
    case PROP_INTER_MASK:
	self->interaction_mask = g_value_get_int(value);
	if (self->bbox_overlay)
	    g_object_set(self->bbox_overlay,
			 "handles", self->interaction_mask, NULL);

	break;

    case PROP_WHOLE_AREA_INTERACTION:
	self->whole_area_interaction = g_value_get_boolean(value);
	break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }

}



static void
get_property (GObject      *object,
	      guint         prop_id,
	      GValue       *value,
	      GParamSpec   *pspec)
{
    NxpCanvasObject *self;

    self = NXP_CANVAS_OBJECT(object);

    switch (prop_id)
    {
    case PROP_CANVAS:
	if (self->canvas)
	    g_value_set_object(value, G_OBJECT(self->canvas));
	break;

    case PROP_INTER_MASK:
	g_value_set_int(value, self->interaction_mask);
	break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }

}


gboolean nxp_canvas_object_button_press(NxpCanvasObject *self, GdkEvent *ev1)
{
    GdkEventButton *event = (GdkEventButton *) ev1;
    gboolean rv;
    g_return_val_if_fail(self, FALSE);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), FALSE);
    g_signal_emit(G_OBJECT(self), signals[BUTTON_PRESS_SIGNAL], 0, event, &rv);
    return rv;
}


static gboolean nxp_canvas_object_real_button_press(NxpCanvasObject *self, GdkEvent *ev1)
{
    GdkEventButton *event = (GdkEventButton *) ev1;
    
/*     g_message(__FUNCTION__); */

    if (self->interaction != NXP_COBJ_INTERACTION_NONE) return TRUE;
    self->interaction = nxp_canvas_object_get_interaction_at_point
	(self, event->x, event->y);
    if (self->interaction == NXP_COBJ_INTERACTION_NONE) {
	/* this shouldn't be needed, but... */
	nxp_canvas_container_pointer_ungrab(self->container);
	nxp_canvas_overlay_hide(self->ghost_bbox);
	return FALSE;
    }
    nxp_canvas_bbox_set_geometry(NXP_CANVAS_BBOX(self->ghost_bbox),
				 &self->wbbox);
    nxp_canvas_overlay_show(self->ghost_bbox);
    /* Don't show handles during move/resize */
    g_object_set(self->bbox_overlay, "handles", 0, 
		 "line-style", GDK_LINE_SOLID,
		 NULL);
    nxp_canvas_container_pointer_grab(self->container, self, TRUE);
    self->drag_origin.x = event->x;
    self->drag_origin.y = event->y;
    return TRUE;
}


gboolean nxp_canvas_object_button_release(NxpCanvasObject *self, GdkEvent *ev1)
{
    GdkEventButton *event = (GdkEventButton *) ev1;
    gboolean rv;
    g_return_val_if_fail(self, FALSE);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), FALSE);
    g_signal_emit(G_OBJECT(self), signals[BUTTON_RELEASE_SIGNAL], 0, event, &rv);
    return rv;
}


static gboolean nxp_canvas_object_real_button_release(NxpCanvasObject *self, GdkEvent *ev1)
{
    GdkEventButton *event = (GdkEventButton *) ev1;
    if (self->interaction != NXP_COBJ_INTERACTION_NONE) {
	GdkRectangle wrect;
	NxplotRect srect;
	NxpCanvasObjectInteraction inter;
	nxp_canvas_object_get_interaction_bbox(self, self->interaction,
					       event->x, event->y,
					       &wrect);
	nxplot_rect_window2space(self->canvas->plotter, &wrect, &srect);
	wrect.x += WBBOX_BORDER;
	wrect.y += WBBOX_BORDER;
	wrect.width  -= WBBOX_BORDER*2;
	wrect.height -= WBBOX_BORDER*2;
	inter = self->interaction;
	self->interaction = NXP_COBJ_INTERACTION_NONE;
	nxp_canvas_overlay_hide(self->ghost_bbox);
	g_object_set(self->bbox_overlay,
		     "handles", self->interaction_mask,
		     "line-style", GDK_LINE_ON_OFF_DASH,
		     NULL);
	nxp_canvas_container_pointer_ungrab(self->container);
	nxp_canvas_object_configure(self, &srect, inter);
	nxp_canvas_container_check_mouse_over(self->container, event->x, event->y);
	return TRUE;
    }
    return FALSE;
}


gboolean nxp_canvas_object_motion_notify(NxpCanvasObject *self, GdkEvent *ev1)
{
    GdkEventMotion *event = (GdkEventMotion *) ev1;
    gboolean rv;
    g_return_val_if_fail(self, FALSE);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), FALSE);
    g_signal_emit(G_OBJECT(self), signals[MOTION_NOTIFY_SIGNAL], 0, event, &rv);
    return rv;
}


static gboolean nxp_canvas_object_real_motion_notify(NxpCanvasObject *self,
						     GdkEvent *ev1)
{
    GdkEventMotion *event = (GdkEventMotion *) ev1;
    if (self->interaction != NXP_COBJ_INTERACTION_NONE) {
	GdkRectangle tmp_geometry;
	nxp_canvas_object_get_interaction_bbox(self, self->interaction,
					       event->x, event->y,
					       &tmp_geometry);
	nxp_canvas_overlay_hide(self->bbox_overlay);
	nxp_canvas_bbox_set_geometry(NXP_CANVAS_BBOX(self->bbox_overlay),
				     &tmp_geometry);
	nxp_canvas_overlay_show(self->bbox_overlay);
    } else  if (self->interaction == NXP_COBJ_INTERACTION_NONE) {
	if (self->interaction_mask != 0)
	    nxp_canvas_object_set_interaction_cursor
		(self, nxp_canvas_object_get_interaction_at_point
		 (self, event->x, event->y));
    }
    return TRUE;
}


#define HANDLE_SIZE 6

/**
 * nxp_canvas_object_get_interaction_at_point:
 * @self: object
 * @x: pointer x corrdinate
 * @y: pointer y corrdinate
 * 
 * Get the kind of interaction will occur if the user drags with
 * button 1 at the given coordinates.
 * 
 * Return value: The interaction type ( =NXP_COBJ_INTERACTION_NONE if
 * no interaction would occur)
 **/
NxpCanvasObjectInteraction
nxp_canvas_object_get_interaction_at_point(NxpCanvasObject *self, gint x, gint y)
{
    gint x1, x1in, x2in, x2, xm1, xm2;
    gint y1, y1in, y2in, y2, ym1, ym2; 
    NxpCanvasObjectInteraction default_interaction;

    g_return_val_if_fail(self, NXP_COBJ_INTERACTION_NONE);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), NXP_COBJ_INTERACTION_NONE);

    default_interaction = self->interaction_mask & NXP_COBJ_INTERACTION_MOVE_MASK;

    x1 = self->wbbox.x;
    x1in = x1 + HANDLE_SIZE - 1;
    x2 = self->wbbox.x + self->wbbox.width - 1;
    x2in = x2 - HANDLE_SIZE + 1;
    xm1 = (x1 + x2)/2 - HANDLE_SIZE/2;
    xm2 = (x1 + x2)/2 + HANDLE_SIZE/2;
    y1 = self->wbbox.y;
    y1in = y1 + HANDLE_SIZE - 1;
    y2 = self->wbbox.y + self->wbbox.height - 1;
    y2in = y2 - HANDLE_SIZE + 1;
    ym1 = (y1 + y2)/2 - HANDLE_SIZE/2;
    ym2 = (y1 + y2)/2 + HANDLE_SIZE/2;

    if (x >= x1 && x <= x1in) {
	if (y >= y1 && y <= y1in)
	    return (NXP_COBJ_INTERACTION_RESIZE_NW &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_NW_MASK)) ||
		default_interaction;
	else if (y >= y2in && y <= y2)
	    return (NXP_COBJ_INTERACTION_RESIZE_SW &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_SW_MASK)) ||
		default_interaction;
	else if (y >= ym1 && y <= ym2)
	    return (NXP_COBJ_INTERACTION_RESIZE_W &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_W_MASK)) ||
		default_interaction;
	else 
	    return NXP_COBJ_INTERACTION_MOVE &&
		(self->interaction_mask & NXP_COBJ_INTERACTION_MOVE_MASK);
    } else if (x >= x2in && x <= x2) {
	if (y >= y1 && y <= y1in)
	    return (NXP_COBJ_INTERACTION_RESIZE_NE &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_NE_MASK)) ||
		default_interaction;
	else if (y >= y2in && y <= y2)
	    return (NXP_COBJ_INTERACTION_RESIZE_SE &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_SE_MASK)) ||
		default_interaction;
	else if (y >= ym1 && y <= ym2)
	    return (NXP_COBJ_INTERACTION_RESIZE_E &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_E_MASK)) ||
		default_interaction;
	else 
	    return NXP_COBJ_INTERACTION_MOVE &&
		(self->interaction_mask & NXP_COBJ_INTERACTION_MOVE_MASK);
    } else if (y >= y1 && y <= y1in) {
	if (x >= xm1 && x <= xm2)
	    return (NXP_COBJ_INTERACTION_RESIZE_N &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_N_MASK)) ||
		default_interaction;
	else 
	    return NXP_COBJ_INTERACTION_MOVE &&
		(self->interaction_mask & NXP_COBJ_INTERACTION_MOVE_MASK);
    } 
    else if (y >= y2in && y <= y2) {
	if (x >= xm1 && x <= xm2)
	    return (NXP_COBJ_INTERACTION_RESIZE_S &&
		    (self->interaction_mask &
		     NXP_COBJ_INTERACTION_RESIZE_S_MASK)) ||
		default_interaction;
	else 
	    return NXP_COBJ_INTERACTION_MOVE &&
		(self->interaction_mask & NXP_COBJ_INTERACTION_MOVE_MASK);
    }
    if (self->whole_area_interaction)
	return NXP_COBJ_INTERACTION_MOVE;
    else
	return NXP_COBJ_INTERACTION_NONE;
}


void nxp_canvas_object_set_interaction_cursor(NxpCanvasObject *self,
					      NxpCanvasObjectInteraction inter)
{
    static GdkCursor *cursors[10] = { NULL, NULL, NULL, NULL, 
				      NULL, NULL, NULL, NULL, NULL, NULL };
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));

    g_return_if_fail(inter >= NXP_COBJ_INTERACTION_NONE &&
		     inter <= NXP_COBJ_INTERACTION_RESIZE_NW);
    
    
    if (!cursors[1]) {
	cursors[NXP_COBJ_INTERACTION_MOVE] = 
	    gdk_cursor_new(GDK_HAND2);
	cursors[NXP_COBJ_INTERACTION_RESIZE_N] =
	    gdk_cursor_new(GDK_BASED_ARROW_UP);
	cursors[NXP_COBJ_INTERACTION_RESIZE_S] =
	    gdk_cursor_new(GDK_BASED_ARROW_DOWN);
	cursors[NXP_COBJ_INTERACTION_RESIZE_E] =
	    gdk_cursor_new(GDK_SB_RIGHT_ARROW);
	cursors[NXP_COBJ_INTERACTION_RESIZE_W] =
	    gdk_cursor_new(GDK_SB_LEFT_ARROW);
	cursors[NXP_COBJ_INTERACTION_RESIZE_NW] =
	    gdk_cursor_new(GDK_TOP_LEFT_CORNER);
	cursors[NXP_COBJ_INTERACTION_RESIZE_NE] =
	    gdk_cursor_new(GDK_TOP_RIGHT_CORNER);
	cursors[NXP_COBJ_INTERACTION_RESIZE_SW] =
	    gdk_cursor_new(GDK_BOTTOM_LEFT_CORNER);
	cursors[NXP_COBJ_INTERACTION_RESIZE_SE] =
	    gdk_cursor_new(GDK_BOTTOM_RIGHT_CORNER);
    }

    nxp_canvas_set_cursor(self->canvas, cursors[inter]);
}


void nxp_canvas_object_get_interaction_bbox(NxpCanvasObject *self, 
					    NxpCanvasObjectInteraction inter,
					    gint x, gint y,
					    GdkRectangle *new_bbox)
{
    GdkPoint delta;
    delta.x = x - self->drag_origin.x;
    delta.y = y - self->drag_origin.y;
    new_bbox->x = self->wbbox.x;
    new_bbox->y = self->wbbox.y;
    new_bbox->width = self->wbbox.width;
    new_bbox->height = self->wbbox.height;

    switch (inter) 
    {
    case NXP_COBJ_INTERACTION_NONE:
	return;
    case NXP_COBJ_INTERACTION_MOVE:
	new_bbox->x += delta.x;
	new_bbox->y += delta.y;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_N:
	new_bbox->y += delta.y;
	new_bbox->height -= delta.y;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_S:
	new_bbox->height += delta.y;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_W:
	new_bbox->x += delta.x;
	new_bbox->width -= delta.x;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_E:
	new_bbox->width += delta.x;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_NW:
	new_bbox->x += delta.x;
	new_bbox->width -= delta.x;
	new_bbox->y += delta.y;
	new_bbox->height -= delta.y;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_NE:
	new_bbox->width += delta.x;
	new_bbox->y += delta.y;
	new_bbox->height -= delta.y;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_SE:
	new_bbox->width += delta.x;
	new_bbox->height += delta.y;
	return;
    case NXP_COBJ_INTERACTION_RESIZE_SW:
	new_bbox->x += delta.x;
	new_bbox->width -= delta.x;
	new_bbox->height += delta.y;
	return;
    case NXP_COBJ_INTERACTION_MOVE_RESIZE:
	g_assert_not_reached();
    }
}



void nxp_canvas_object_configure(NxpCanvasObject *self, NxplotRect *new_bbox,
				 NxpCanvasObjectInteraction inter)
{
    NxplotRect old_bbox;

    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));

    old_bbox = self->bbox;
    g_signal_emit(G_OBJECT(self), signals[CONFIGURE_SIGNAL], 0, new_bbox, inter);
    if (inter == NXP_COBJ_INTERACTION_MOVE_RESIZE) {
	self->bbox = *new_bbox;
	if (g_type_is_a(G_OBJECT_TYPE(self->canvas->plotter), TYPE_NXPLOT_ARTFT))
	{
/* 	    g_message("%p bbox: (%G, %G)-(%G, %G) size %G x %G", self, */
/* 		      self->bbox.x1, self->bbox.y1, */
/* 		      self->bbox.x2, self->bbox.y2, */
/* 		      self->bbox.x2 - self->bbox.x1, */
/* 		      self->bbox.y2 - self->bbox.y1); */
	    nxp_canvas_object_update_wbbox(self);
	    nxp_canvas_redraw_area(self->canvas, &old_bbox);
	    nxp_canvas_redraw_area(self->canvas, &self->bbox);
	}
    }
}


NxplotPoint* nxp_canvas_object_size_request(NxpCanvasObject *self)
{
    NxplotPoint *rv = NULL;
    g_return_val_if_fail(self, NULL);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), NULL);

    if (!self->canvas) return NULL;
    if (!self->canvas->plotter) return NULL;

    if (g_type_is_a(G_OBJECT_TYPE(self->canvas->plotter), TYPE_NXPLOT_ARTFT)) {
	if (!nxplot_artft_ready(NXPLOT_ARTFT(self->canvas->plotter)))
	    return NULL;
    }

    g_signal_emit(G_OBJECT(self), signals[SIZE_REQUEST_SIGNAL], 0, &rv);
    return rv;
}


gboolean nxp_canvas_object_key_press(NxpCanvasObject *self, GdkEvent *ev1)
{
    GdkEventKey *event = (GdkEventKey *) ev1;
    gboolean rv;
    g_return_val_if_fail(self, FALSE);
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), FALSE);
    g_signal_emit(G_OBJECT(self), signals[KEY_PRESS_SIGNAL], 0, event, &rv);
    return rv;
}


void nxp_canvas_object_update(NxpCanvasObject *self)
{
    g_return_if_fail(self);
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));

    nxp_canvas_object_queue_draw(self);
    g_signal_emit(G_OBJECT(self), signals[UPDATE_SIGNAL], 0);
    nxp_canvas_object_queue_draw(self);
}


/* Don't use this! This class is really abstract, but this function is
 * needed for language bindings. */
NxpCanvasObject* nxp_canvas_object_new(void)
{
    return g_object_new(NXP_TYPE_CANVAS_OBJECT, NULL);
}


static void container_canvas_notify(NxpCanvasContainer *container,
				    GParamSpec         *pspec,
				    NxpCanvasObject    *obj)
{
    NxpCanvas *canvas = NULL;
    g_return_if_fail(NXP_IS_CANVAS_CONTAINER(container));
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(obj));
    
    g_object_get(container, "canvas", &canvas, NULL);
    nxp_canvas_object_set_canvas(obj, canvas);
}


void nxp_canvas_object_set_container(NxpCanvasObject *self, NxpCanvasContainer *container)
{
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    g_return_if_fail(NXP_IS_CANVAS_CONTAINER(container));
    g_return_if_fail(!self->container);
    if (NXP_CANVAS_OBJECT(container)->canvas && self->canvas && 
	self->canvas != NXP_CANVAS_OBJECT(container)->canvas)
	g_warning("%s: object belongs to canvas different from container's", __FUNCTION__);
    self->container = container;
    if (NXP_CANVAS_OBJECT(container)->canvas)
	nxp_canvas_add_object(NXP_CANVAS_OBJECT(container)->canvas, self);
    else {
	if (self->canvas_notify_sig)
	    g_signal_handler_disconnect(self->container, self->canvas_notify_sig);
	self->canvas_notify_sig =
	    g_signal_connect(container, "notify::canvas",
			     G_CALLBACK(container_canvas_notify), self);
    }
}


void nxp_canvas_object__set_visible(NxpCanvasObject *obj, gboolean visible)
{
    g_return_if_fail(obj);
    obj->visible = visible;
    if (obj->canvas)
	nxp_canvas_recompute_drawable_objects_list(obj->canvas);
}


void nxp_canvas_object__set_drawable(NxpCanvasObject *obj, gboolean drawable)
{
    g_return_if_fail(obj);
    obj->drawable = drawable;
}

void nxp_canvas_object_set_canvas(NxpCanvasObject *self, NxpCanvas *canvas)
{
    g_return_if_fail(NXP_IS_CANVAS_OBJECT(self));
    g_return_if_fail(NXP_IS_CANVAS(canvas));

    self->canvas = canvas;
    nxp_canvas_add_object(canvas, self);
    if (self->bbox_overlay)
	nxp_canvas_overlay_set_canvas(self->bbox_overlay, canvas);
    if (self->ghost_bbox)
	nxp_canvas_overlay_set_canvas(self->ghost_bbox, canvas);
    g_object_notify(G_OBJECT(self), "canvas");
}


Nxplot* nxp_canvas_object_get_plotter(NxpCanvasObject *self)
{
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), NULL);
    g_return_val_if_fail(self->canvas, NULL);
    return self->canvas->plotter;
}


NxpCanvasContainer* nxp_canvas_object_get_toplevel(NxpCanvasObject *self)
{
    g_return_val_if_fail(NXP_IS_CANVAS_OBJECT(self), NULL);
    NxpCanvasObject *obj = self;

    while (obj->container)
	obj = NXP_CANVAS_OBJECT(obj->container);
    return obj != self? NXP_CANVAS_CONTAINER(obj) : NULL;
}


